"""
         Yasuharu Okamoto, 2024.08
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
#import japanize_matplotlib
from scipy.cluster.hierarchy import linkage
from scipy.cluster.hierarchy import dendrogram
 
#         Read CSV file
fl_nm = input('CSV File(*.csv) = ')
df = pd.read_csv(fl_nm, encoding='shift-jis')

#       Names of column items 
col_names = df.keys()    
col_names = col_names[1:]   
#        Names of row items
row_names = df.values.T[0]
#
#       Values in the two dimensional array data
#
df_data = df.values    
F = df_data[:,1:]          
data = np.array(F, dtype=float)

print(data)
#       Singular Value Decomposition of data 
u, s, vh = np.linalg.svd(data, full_matrices = False)
p = sum([1 if v > 1.0E-9 else 0 for v in s])
s = s[:p]
print('\ns =', s)
print('p =', p)
u = u[:,:p] 
vh = vh[:p]
#
#         Scree plot of the singular values
#
plt.title('Scree Plot of Singular Values', fontsize=14)
plt.plot(range(1, p+1), s, c = 'k')
plt.xticks(np.arange(1, p+1, 1))
plt.ylabel('Value', fontsize = 14)
plt.tight_layout()
plt.show()
plt.close()
#
#            Weighting
#
M = len(data)
N = len(data.T)
s1 = np.empty(p)
s2 = np.empty(p)
for t in range(p):
    s1[t] = (s[t]**0.5) * ((M/N)**0.25)  
    s2[t] = (s[t]**0.5) * ((N/M)**0.25)  

row_coord = u @ np.diag(s1)
col_coord = np.diag(s2) @ vh

#
#          Concatenation of the row and column coordinates
#
mrg_coord = np.concatenate((row_coord, col_coord.T), axis = 0)
mrg_label = np.concatenate((row_names, col_names), axis = 0)

#
#           Eigendecomposition with centering
#
mean_X = np.mean(mrg_coord, axis = 0)
Z = mrg_coord - mean_X
w, v = np.linalg.eigh(Z.T @ Z)

sidx = np.argsort(-w)
w = w[sidx]
v = v.T[sidx].T
#      Scree plot of the eigen values
plt.title('Scree Plot for w', fontsize=16)
plt.plot(np.arange(len(w))+1,w, c='k')
plt.xticks(np.arange(len(w))+1)
plt.ylabel('Eigen value', fontsize=14)
plt.tight_layout()
plt.show()
plt.close()
#
#       Rotation of the coordinates
#
Proj = v.T 
Proj_mrg_coord = (Proj @ mrg_coord.T).T
#
#       Output the coordinates to the file Coord.txt
#
p_dim = 5 if p > 5 else p
fout = open('Coord.txt', 'w')
fout.write(f'\n\n{"Names":<15s}')
for i in range(p_dim):
    fout.write(f'{f"Dim-{i+1}":<15s}')
fout.write('\n')
for i in range(M):
    fout.write(f'{row_names[i]:<15s}')
    for j in range(p_dim):
        fout.write(f'{Proj_mrg_coord[i][j]:<15.5f}')
    fout.write('\n')

fout.write('\n')
for i in range(N):
    fout.write(f'{col_names[i]:<15s}')
    for j in range(p_dim):
        fout.write(f'{Proj_mrg_coord[M+i][j]:<15.5f}')
    fout.write('\n')

fout.close()
print('\n', 'Coord.txt', 'was saved.')

#
#        Plot of coordinates
#
max_coord = np.max(np.abs(Proj_mrg_coord))

dimx = 0
dimy = 1
while True:
    plt.close()
    plt.figure(figsize=(7,7))   
    plt.xticks([-max_coord, max_coord])
    plt.yticks([-max_coord, max_coord])
    #      Plot the row items
    plt.plot(Proj_mrg_coord.T[dimx][:M], Proj_mrg_coord.T[dimy][:M],
             '^', c = 'b', label = 'Row')
    #      Plot the column itens
    plt.plot(Proj_mrg_coord.T[dimx][M:], Proj_mrg_coord.T[dimy][M:],
             'v', c = 'magenta', label = 'Col')
    #      Write the row item names
    for i in range(M):
        plt.text(Proj_mrg_coord.T[dimx][i], Proj_mrg_coord.T[dimy][i] - 0.05,
                 row_names[i], rotation = 0, fontsize = 14, color='b')
    #      Write the column item names
    for i in range(N):
        plt.text(Proj_mrg_coord.T[dimx][M + i], Proj_mrg_coord.T[dimy][M + i] + 0.05,
                 col_names[i], rotation = 0, fontsize = 14, color='magenta')
    plt.plot([0], [0], '*', c = 'k', markersize = 12, label = 'Origin')
    plt.plot([-max_coord, max_coord], [0, 0], c = 'k', ls = ':')
    plt.plot([0, 0], [-max_coord, max_coord], c = 'k', ls = ':')
    plt.legend(loc='upper left')
    plt.title('Orthogonally Rotated Configulation', fontsize = 16)
    plt.xlabel(f'Axis-{dimx+1}', fontsize = 14)
    plt.ylabel(f'Axis-{dimy+1}', fontsize = 14)
    plt.tight_layout()
    plt.show()
    #
    #      Selection of the axes for the next plot 
    #
    print(f'\nMaximum dimension = {p}')
    dimx = int(input('Dimension for x-axis = '))
    if dimx < 1:    #    End this plotting 
        break
    if dimx > p:
        dimx = p
    dimx -= 1
    dimy = int(input('Dimension for y-axis = '))
    if dimy < 1:    #     End this plotting
        break
    if dimy > p:
        dimy = p
    dimy -= 1

#
#     Concatenate the row and column coordinates for clustering
#
mrg_coord = np.concatenate((row_coord, col_coord.T), axis = 0)
#
#     Concatenate the row and column names for clustering
#
mrg_label = np.concatenate((row_names, col_names), axis = 0)
#
#        Clustering by the coordinates
#
row_clusters = linkage(mrg_coord, method = 'ward')
fig = plt.figure(figsize = (7, 5), facecolor = 'white')
row_dendr = dendrogram(row_clusters, labels = mrg_label,
                       orientation = 'right', leaf_font_size = 14,
                       link_color_func = lambda k: 'k')
plt.title('Clustering by Coordinates', fontsize=14)
plt.tight_layout()
plt.show()

#
#      Calculate the angle between two vectors
#
def calc_angle(x, y):
    v = np.dot(x, y) / ((np.dot(x,x))*np.dot(y,y))**0.5
    v1 = 1.0 if v > 1.0 else -1.0 if v < -1.0 else v
    return np.arccos(v1)

#
#      Calculation of angles between vectors of items
#

Angles = np.empty((M+N, M+N))
for i in range(M+N):
    for j in range(M+N):
        if i > j:
            Angles[i][j] = Angles[j][i]
        elif i == j:
            Angles[i][j] = 0.0
        else:
            Angles[i][j] = calc_angle(mrg_coord[i], mrg_coord[j])    

#
#       Cluster Analysis by angles
#
dists = []
for i in range(len(Angles)-1):
    for j in range(i+1, len(Angles)):
        dists.append(Angles[i][j])

row_clusters = linkage(dists, method = 'complete')
fig = plt.figure(figsize = (7, 5), facecolor = 'white')
row_dendr = dendrogram(row_clusters, labels = mrg_label,
                       orientation = 'right', leaf_font_size = 14, link_color_func = lambda k: 'k')
plt.title('Clustering by Angles', fontsize = 14)
plt.tight_layout()
plt.show()


